//  GenomeMixer.h

#ifndef __GENOME_MIXER_MAIN_FILE__
#define __GENOME_MIXER_MAIN_FILE__

#define kTABLE_SIZE   200

#include "Mouse.h"
#include "MarkerCollection.h"
#include "MouseCell.h"
#include "CellData.h"
#include "PurebredData.h"

#include <iostream>
#include <string>
#include <vector>
#include <stdlib.h>
#include <time.h>
#include <math.h>

class GenomeMixer {
    private:
        int myNumCages;
        int myNumGenerations;
        vector<Mouse*> miceToExport;	
	// Also, the two 2d-arrays below should really be dynamically allocated.
	string       myBreedingTable[kTABLE_SIZE][kTABLE_SIZE]; // The contents of the breeding scheme table
	MouseCell    myMiceTable[kTABLE_SIZE][kTABLE_SIZE];  // # the big table of cages vs. generation, the actual litters (MouseCells) that go into each cell
	CellData*    myCellDataTable[kTABLE_SIZE][kTABLE_SIZE];
	PurebredData pure;
	
    public:
        GenomeMixer::GenomeMixer(); // constructor
	
	Mouse*      MakeOneMouse(char gender, Mouse *mother, Mouse *father, int litterNumber, int siblingNumber, MarkerCollection *theMarkerCollection);
	MouseCell*  NewLitter(MouseCellIngredients *in, MarkerCollection *theMarkerCollection);
        void      GenomeMixer::clearExportList(void);
	void      GenomeMixer::addExport(Mouse* mousePtr);
	void      GenomeMixer::printListOfExportMice(void);
	void      GenomeMixer::setTableSize(int height, int width);
        
        MouseCell*  GenomeMixer::getMiceTable(int cage, int gen);
                
	void      GenomeMixer::setBreedingTableElem(int cage, int gen, string x);
	string    GenomeMixer::getBreedingTable(int cage, int gen);
		
	void      GenomeMixer::setCellData(int cage, int gen, CellData *theData);
	CellData* GenomeMixer::getCellData(int cage, int gen);
	void      GenomeMixer::clearCellDataTable(void);
	
	// Purebred-related functions...
	PurebredData* GenomeMixer::purePtr(void);	
	void   GenomeMixer::clearAllParentalStrains(void);
	void   GenomeMixer::addParentalStrain(string strainName, string symbol);	
	int    GenomeMixer::numParentalStrains(void);
	void   GenomeMixer::makePurebreds(void);	
	void   GenomeMixer::printParentalStrains(void); // Prints to terminal out
	Mouse* GenomeMixer::getPurebredPtr(int index, bool sex);
	string GenomeMixer::getPurebredSymbol(int index);
	string GenomeMixer::getPurebredName(int index);	
	
	void   GenomeMixer::runSimulation(void);
	
        int    GenomeMixer::numCages(void);
        int    GenomeMixer::numGenerations(void);
	Mouse* GenomeMixer::mouseToExport(int index);
	int    GenomeMixer::numMiceToExport(void);
	
	void GenomeMixer::ChromosomeCombine(Mouse *newMousePtr, MarkerCollection *markerColPtr);
	void GenomeMixer::copyChr(int chrNum, Mouse *destMouseRef, Mouse *sourceMouseRef, bool sourceSide, bool destSide);
	
	Chromosome* GenomeMixer::combineParentalChr(int chrNum, Mouse *parentPtr, MarkerCollection *markerColPtr);
	Chromosome* GenomeMixer::segmentChr(Chromosome *sourceChr, vector<long> *breaksArrayPtr); // (called by combineParentalChr)
	
	vector<long>* GenomeMixer::pickRecombLocs(int thisChrNum, MarkerCollection *markerColPtr);
	
	static void GenomeMixer::printOutRecombLocs(vector<long> *recombLocsPtr, int thisChrNum);
	
	bool GenomeMixer::checkProposedPoint(long thisPoint, vector<long> *vecOfBreakPts);
	bool GenomeMixer::hasARecomb(float cM_dist);
	
			
	// Below: Public instance variables
		
	MarkerCollection *markerColPtr;
	MouseCell *mice;
        	
	int    defaultLitterSize; // Default size for a new litter of mice
        int    defaultLitterStDev; // Default standard deviation for the size of a litter (not yet used)
        string name;    // The name of this GenomeMixer object
	float  version; // version number
                
	long   noRecombDistBases; // absolute interference
};


#endif
		
