// ================================================================
// GenomeMixer    (c) 2003, Alex Williams
// All rights reserved.
// Send email correspondence to Robert W. Williams <rwilliam@nb.utmem.edu>
// at the University of Tennesse Health Science Center

// This project was funded jointly by the National Institute of Mental Health, National Institute on Drug Abuse,
// and the National Science Foundation (award P20-MH 62009 to KFM and RW)

// This source code may be used according to the terms of the GNU General Public License (GPL).
// See www.gnu.org for specifics. Essentially, the source may be freely used as long as the source code
// for whatever program the source is also made freely available. If you want to use this source
// in a closed-source program, you should email us for licensing options.

// Disclaimer: This code comes with no warranty at all, and there is no guarantee that it will
// work as claimed in the documentation. Hopefully, it will be useful anyway.

// Note that GenomeMixer makes use of the Trolltech Qt toolkit for its window-managing needs.
// There is very little use of Qt data structures (aside from QString) in the actual processing code,
// but **you will need to have Qt installed in order to compile this code.** Check <www.trolltech.com>
// for details; there should be a free GPL-software-only version of Qt available on the Trolltech
// web site.

// Any redistributions of this software in code or binary form must include the
// above copyright notice, licensing information, disclaimer, and this notice itself.
// Basically, all the comments at the top of this file must be retained.
// ================================================================


/****************************************************************************
** ui.h extension file, included from the uic-generated form implementation.
**
** If you wish to add, delete or rename functions or slots use
** Qt Designer which will update this file, preserving your code. Create an
** init() function in place of a constructor, and a destroy() function in
** place of a destructor.
*****************************************************************************/

const QString globalVersionStr("0.53a");
const QString globalDateStr("August 2003");

using namespace std;

void MixerApp::init() {    
	mainMixer = new GenomeMixer();
	
	outputQGene = false;
	outputMapManager = false;
	
	settingsNeedSaving = false;
	markersNeedSaving = false;
	
	updateGeneralInfo();
	
	textGeneralInfo->setText("Welcome to GenomeMixer " + globalVersionStr + ". See the About box for version\ninformation, and the information menu for help.");
}


void MixerApp::fileNewSettings() {
	if (settingsNeedSaving) {
		// warn the user about unsaved changes first
		switch(QMessageBox::warning(this, "GenomeMixer: Unsaved Changes",
									"If you create a new settings file before saving the\n"
									"old settings, then the unsaved settings will be lost.",
									"New Settings Anyway", "Cancel (don't discard)", QString::null, 1, 1)) {
			case 0: // discard changes: continue onward, not bothering to save
			return;
			break;
			case 1: // cancel
			return; // go back, giving the user an opportunity to save
			break;
		}
	}
	
	settingsNeedSaving = true;
	settingsLocation = QString::null; // <-- null string
	
	MarkerCollection *tempCol = NULL; // <
	if (mainMixer && mainMixer->markerColPtr) {
		tempCol = mainMixer->markerColPtr;
	}
	if (mainMixer) {
		delete mainMixer;
	}
	mainMixer = new GenomeMixer();
	mainMixer->markerColPtr = tempCol;
	
	updateGeneralInfo();
}




void MixerApp::fileNewMarkers() {
	if (markersNeedSaving) {
		// warn the user about unsaved changes first
		switch(QMessageBox::warning(this, "GenomeMixer: Unsaved Changes",
									"If you make a new marker file now,\n"
									"some unsaved marker changes will be lost.",
									"New Markers Anyway", "Cancel (go back)", QString::null, 1, 1)) {
		case 0:
			// discard changes: continue onward, not bothering to save
			break;
			case 1: // cancel
			return; // go back, giving the user an opportunity to save 
			break;
		}
	}
	
	markersNeedSaving = true;
	markersLocation = QString::null; // <-- null string
	
	if (mainMixer && mainMixer->markerColPtr) {
		delete mainMixer->markerColPtr;
	}
	mainMixer->markerColPtr = new MarkerCollection(0);
	
	updateGeneralInfo();
}



void MixerApp::updateGeneralInfo() {
	
	if (settingsLocation.isNull())	{
		// in other words, we don't have a settingsLocation file
		if (settingsNeedSaving)
			labelFileToOpen->setText(QString("(new settings: not saved to disk yet)"));
		else
			labelFileToOpen->setText(QString("(none selected)"));
	} else {
		labelFileToOpen->setText(settingsLocation);
	}
	
	if (markersLocation.isNull()) {
		if (markersNeedSaving) 
			labelMarkerFile->setText(QString("(new markers: not saved to disk yet)"));
		else
			labelMarkerFile->setText(QString("(none selected)"));
	} else {
		labelMarkerFile->setText(markersLocation);
	}
	
	buttonSaveSettings->setEnabled(settingsNeedSaving); // enable/disable the "save settings" button
	buttonSaveMarkers->setEnabled(markersNeedSaving); // enable/disable the "save markers" button
	
	if (settingsNeedSaving) {
		fileMenu->changeItem(fileMenu->idAt(1), "&New Settings...");
		// the ellipsis needs to go there, to indicate that it will pop up a "you really want a new file without saving?" dialog
	} else {
		fileMenu->changeItem(fileMenu->idAt(1), "&New Settings");
		// no file pops up, so no ellipsis
	}
	
	if (markersNeedSaving) {
		fileMenu->changeItem(fileMenu->idAt(2), "New &Marker File...");
		// the ellipsis needs to go there, to indicate that it will pop up a "you really want a new file without saving?" dialog
	} else {
		fileMenu->changeItem(fileMenu->idAt(2), "New &Marker File");
		// no file pops up, so no ellipsis
	}
	
	
	fileMenu->setItemEnabled(fileMenu->idAt(5), settingsNeedSaving); // <-- enables/disables the "Save settings" menu item
	fileMenu->setItemEnabled(fileMenu->idAt(8), markersNeedSaving); // <-- enables/disables the "Save markers" menu item
	
	bool haveMiceToExport = false;
	if (mainMixer->numMiceToExport() > 0) {
		// If we don't have any mice to export, then we don't have any reason to
		// enable the output buttons or the Inspect Progeny button.
		haveMiceToExport = true;
	}
	buttonOutputQGene->setEnabled(haveMiceToExport); // enables/disables the export option buttons
	buttonOutputMM->setEnabled(haveMiceToExport);
	buttonOutputRQtl->setEnabled(haveMiceToExport);	
	outputMenu->setItemEnabled(outputMenu->idAt(0), haveMiceToExport); // enables/disables the export option menu items
	outputMenu->setItemEnabled(outputMenu->idAt(1), haveMiceToExport);
	outputMenu->setItemEnabled(outputMenu->idAt(2), haveMiceToExport);
	
	
	buttonInspectMice->setEnabled(haveMiceToExport); // enables/disables the "inspect progeny" button	
	simulationMenu->setItemEnabled(simulationMenu->idAt(2), haveMiceToExport); // enabled/disables the "inspect progeny" menu item
	
}


void MixerApp::fileOpen() {
	const char marker_file = 0;
	const char genomeMixer_file = 1;
	int whatWeAreReading = -1;
	
	if (settingsNeedSaving || markersNeedSaving) {
		// warn the user about unsaved changes first
		switch(QMessageBox::warning(this, "GenomeMixer: Unsaved Changes",
									"If you open a new file before saving,\n"
									"the unsaved information will be lost.",
									"Open Anyway", "Cancel (don't discard)", QString::null, 1, 1)) {
		case 0:
			// discard changes: continue onward, not bothering to save
			break;
			case 1: // cancel
			return; // go back, giving the user an opportunity to save
			break;
		}
	}
	
	QString theNewFileStr = QFileDialog::getOpenFileName(QString::null, "All Files (*);;GenomeMixer Settings (*.gmx);;Marker File (*.mar)", this, "open it", "Select a GenomeMixer file to open"); // openDialog.getOpenFileName();
	
	if (theNewFileStr.isNull()) {
		return; // not valid, because we didn't actually select a file
	}
	
	if (theNewFileStr.right(4) == ".mar") {
		whatWeAreReading = marker_file;
		//qDebug("Reading a marker file...");
		
	} else if (theNewFileStr.right(4) == ".gmx") {
		whatWeAreReading = genomeMixer_file;
		//qDebug("Reading a GenomeMixer file.");
	}
	
	if (genomeMixer_file == whatWeAreReading) {
		try {
			MouseIO::readBreedingSchemeFile(theNewFileStr, mainMixer);
			settingsNeedSaving = false;
			settingsLocation = theNewFileStr; // Now we should get the settings from this file
			textGeneralInfo->setText("Successfully loaded general settings file\n\"" 
									 + settingsLocation + "\"");
		} catch ( const string *e) {
			QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
			textGeneralInfo->setText("Failed to read the selected settings file.\n");
		}
		
	} else if (marker_file == whatWeAreReading) {
		try {
			if (mainMixer->markerColPtr != NULL)  delete mainMixer->markerColPtr; // <-- NOT "delete[ ]"
			
			mainMixer->markerColPtr = MouseIO::getMarkerDataFromFile(theNewFileStr);
			markersNeedSaving = false;	    
			markersLocation = theNewFileStr; // Now we should update the file location	    
			textGeneralInfo->setText("Successfully loaded marker file\n\"" + markersLocation + "\"");
			//mainMixer->markerColPtr->printMarkerInfo(); // <-- For debugging, prints what we just read in
			
		} catch ( const string *e) {
			QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
			textGeneralInfo->setText("Failed to read the selected marker file.\n");
			mainMixer->markerColPtr = NULL;
		}
		
	} else {
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "We couldn't determine the what kind of file\n"
							 "you selected. The only way we can detect files\n"
							 "is to look at the suffex. So make sure that the\n"
							 "GenomeMixer settings file ends in \".gmx\", and that\n"
							 "the marker file ends in \".mar\".",
							 "Ok");
		return;
	}
	
	updateGeneralInfo();
}



void MixerApp::fileSaveSettings() {
	if (settingsLocation.isNull()) {    // if we haven't specified the save location yet...
		fileSaveSettingsAs(); // <-- we have to specify where to save the file
		
	} else {     // otherwise, try to save it in the same location where we saved previously
		try {
			MouseIO::writeSettingsFile(settingsLocation, mainMixer, "\n"); // \n: UNIX linebreak
			settingsNeedSaving = false;
			textGeneralInfo->setText("Successfully saved settings file\n\"" + settingsLocation + "\"");
		} catch (const string *e) {
			QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
		}
		updateGeneralInfo();
	}
}

void MixerApp::fileSaveSettingsAs() {
	// The markers location changes in this case
	settingsLocation = QFileDialog::getSaveFileName(QString::null, "All Files (*)", this, "saving dialog", "Save the GenomeMixer settings");
	if (settingsLocation.isNull()) {  // <-- user cancelled the save
		textGeneralInfo->setText("Cancelled the save:\nno settings were written.");
		return;
	}
	
	try {
		MouseIO::writeSettingsFile(settingsLocation, mainMixer, "\n"); // \n: UNIX linebreak
		settingsNeedSaving = false;
		textGeneralInfo->setText("Successfully saved settings file\n\"" + settingsLocation + "\"");
	} catch (const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
	}
	updateGeneralInfo();
}




void MixerApp::fileSaveMarkers() {
	if (markersLocation.isNull()) {   // if we haven't specified the save location yet...
		fileSaveMarkersAs(); // <-- we have to specify where to save the file
		
	} else {   // otherwise, try to save it in the same location where we saved previously
		try {
			MouseIO::writeMarkerFile(markersLocation, mainMixer, "\n"); // \n: UNIX linebreak
			markersNeedSaving = false;
			textGeneralInfo->setText("Successfully saved marker file\n\"" + markersLocation + "\"");
		} catch (const string *e) {
			QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
		}
		updateGeneralInfo();
	}
}

void MixerApp::fileSaveMarkersAs() {
	// The markers location changes in this case
	markersLocation = QFileDialog::getSaveFileName(QString::null, "All Files (*)", this, "saving dialog", "Save the GenomeMixer markers");
	if (markersLocation.isNull()) {  // <-- user cancelled the save
		textGeneralInfo->setText("Cancelled the save:\nno markers were written.");
		return;
	}
	try {
		MouseIO::writeMarkerFile(markersLocation, mainMixer, "\n"); // \n: UNIX linebreak
		markersNeedSaving = false;
		textGeneralInfo->setText("Successfully saved marker file\n\"" + markersLocation + "\"");
	} catch (const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
	}
	updateGeneralInfo();
}





void MixerApp::fileQuit() {
	if (settingsNeedSaving || markersNeedSaving) {
		// warn the user about unsaved changes first
		switch(QMessageBox::warning(this, "GenomeMixer: Unsaved Changes",
									"You have unsaved changes, and if you quit\n" 
									"before saving them, they will be lost.",
									"Quit Without Saving", "Cancel (don't quit)", QString::null,
									1, 1)) {
			case 0: // discard changes: continue onward, not bothering to save
			
			break;
			case 1: // cancel
			return; // go back, giving the user an opportunity to save
			break;
		}	
	}    	
	
	//QApplication::quit();
	//this->quit();
	//setDialog.close();
	this->close();
}


void MixerApp::editUndo() {}
void MixerApp::editRedo() {}
void MixerApp::editCut() {}
void MixerApp::editCopy() {}
void MixerApp::editPaste() {}


void MixerApp::openChrSettingsDialog() {
	
	setDialog.setUpSettings(mainMixer);
	
	//setDialog.show(); // <-- makes it "modeless" (just another window, essentially)
	setDialog.exec(); // <-- modal dialog (can't do anything else while the dialog is open)
	/// <-- execution halts here while we wait for the modal dialog to finish and return
	
	markersNeedSaving = setDialog.markersChanged();
	settingsNeedSaving = setDialog.settingsChanged();
	
	updateGeneralInfo();
}





void MixerApp::openMouseInspector()
{
	// The mouse inspector is a read-only inspector of the currently-calculated mice to be exported (or perhaps all of them)
	inspector.setUpInspector(mainMixer);
	inspector.exec(); // or perhaps .show() would be better
}



void MixerApp::callRunSimulation() {
	time_t  startTime; // <-- This gives time in seconds (time_t is a long int, usually, also)
	time(&startTime); // "startTime = current time"
	srand(startTime); // initialize the random seed
	
	textGeneralInfo->setText(QString("Running simulation...\nPlease wait... " ));
	
	if (mainMixer == NULL || mainMixer->numParentalStrains() < 1) {
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "Warning: Cannot run the simulation:\n"
							 "You must define at least one purebred parental\n"
							 "strain in order to run a simulation.\n"
							 "Ok, I need a purebred strains");
		return;
	} else if (mainMixer->markerColPtr == NULL || mainMixer->markerColPtr->totalMarkers() <= 0) {
		QMessageBox::warning(this, "GenomeMixer Alert",
							 "Error: You don't have any markers defined,\n"
							 "and since we require recombinations to happen\n"
							 "between markers, the simulation can't proceed.\n",
							 "Ok, I should add some markers");
		return;
	}
	
	try {
		Mouse::resetID_Counter(); // first,  reset the IDs so we don't keep incrementing...
		mainMixer->makePurebreds(); // <-- then make the purebred strains (note that they get IDs, too)
		mainMixer->runSimulation(); // <-- this is where all of the actual processing takes place (in GenomeMixer.cpp)
	} catch (const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
		
		textGeneralInfo->setText(QString("Simulation halted on an error.\nNo results were produced."));		return; // <-- we return here, so we don't execute the "set running time" below
	}
	
	time_t finishTime; // time(NULL) (returns time, but doesn't put it anywhere)
	time(&finishTime);
	long timeElapsed = finishTime - startTime;
	QString timeToRunStr;
	if (timeElapsed < 1) {
		timeToRunStr = "< 1 second";
	} else if (timeElapsed >= 1 && timeElapsed < 2) {
		timeToRunStr = "1 second";
	} else {
		timeToRunStr = QString::number(timeElapsed) + " seconds";
	}
	textGeneralInfo->setText(QString("Finished with simulation.\nRunning time: " ) + timeToRunStr);
	
	updateGeneralInfo();
}



void MixerApp::saveQGeneFormat(void) {
	try {
		
		if (mainMixer && mainMixer->numParentalStrains() > 2) {
			// Let the user know that the current output to QGene only works properly with 2 parental strains (you can have more in the file, but only two can be used, if you want the information to be accurate)
			QMessageBox::information(this, "GenomeMixer",
									 "Note: QGene output is currently only compatible with\n crosses that originate from AA and BB parental strains.\nIf you have a CC parental strain,\nthe QGene output will not work properly for mice descended from\nthat strain, although we still export data.\nOne workaround is to do your simulation the normal\nway, but to set all the purebred\nsymbols to 'A' or 'B' in the purebreds tab\nbefore you export.", 
									 "Ok");
		}
		
		QString saveQGeneLocation = QFileDialog::getSaveFileName(QString::null, "All Files (*)", this, "saving dialog", "Save QGene-format output");	
		
		if (saveQGeneLocation.isNull()) {
			textGeneralInfo->setText("Cancelled save.\n ");
		} else {
			MouseIO::writeQGeneFiles(saveQGeneLocation, mainMixer, "\r"); // \r: Mac linebreak
			textGeneralInfo->setText("Successfully wrote QGene files (.pop and .map).\n ");
		}
		
	} catch ( const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
	}
}



void MixerApp::saveMapManagerFormat(void) {
	try {	
		if (mainMixer && mainMixer->numParentalStrains() > 2) {
			// Let the user know that the current output to QGene only works properly with 2 parental strains (you can have more in the file, but only two can be used, if you want the information to be accurate)
			QMessageBox::information(this, "GenomeMixer",
									 "Note: MapManager output is currently only compatible with\n crosses that originate from A and B parental strains.\nIf you have a CC parental strain,\nthe output will not work properly for mice descended from\nthat strain, although we still export data.\nOne workaround is to do your simulation the normal\nway, but to set all the purebred\nsymbols to 'A' or 'B' in the purebreds tab\nbefore you export.", 
									 "Ok");
		}
		
		
		QString saveMMLocation = QFileDialog::getSaveFileName(QString::null, "All Files (*)", this, "saving dialog", "Save MapManagerQTX-format output");
		if (saveMMLocation.isNull()) {
			textGeneralInfo->setText("Cancelled save.\n ");
		} else {
			MouseIO::writeMapManagerFormat(saveMMLocation, mainMixer, "\n"); // \n: UNIX linebreak
			textGeneralInfo->setText("Successfully wrote MapManager-format file.\nLinebreak format was \'\n\' UNIX-style linefeed.");
		}
	} catch ( const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
	}
}

void MixerApp::saveRQtlFormat() {
	try {
		if (mainMixer && mainMixer->numParentalStrains() > 2) {
			// Let the user know that the current output to QGene only works properly with 2 parental strains (you can have more in the file, but only two can be used, if you want the information to be accurate)
			QMessageBox::information(this, "GenomeMixer",
									 "Note: R/qtl output is currently only compatible with\n crosses that originate from A and B parental strains.\nIf you have a third parental strain,\nthe R/qtl output will not work properly for mice descended from\nthat strain, although we still export data.\nOne workaround is to do your simulation the normal\nway, but to set all the purebred\nsymbols to 'A' or 'B' in the purebreds tab\nbefore you export.", 
									 "Ok");
		}
		
		QMessageBox::information(this, "GenomeMixer",
									 "Note: R/qtl may give you a warning about the X chromosome being\nincorrect, if any of the X chromosomes have the 'B'\nallele only (as opposed to 'A' or 'H'). Some R/qtl functions\nwill not work, but things should be okay in general.", 
									 "Ok");
		
		QString saveRQtlLocation = QFileDialog::getSaveFileName(QString::null, "All Files (*)", this, "saving dialog", "Save R/qtl output");	
		if (saveRQtlLocation.isNull()) {
			textGeneralInfo->setText("Cancelled save.\n ");
		} else {
			MouseIO::writeRQtlFormat(saveRQtlLocation, mainMixer, "\n"); // \n: UNIX linebreaks
			textGeneralInfo->setText("Successfully wrote R/qtl-format file.\n ");
		}
	} catch ( const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
	}
}






void MixerApp::showAboutBox()
{
	QMessageBox::about(this, "About GenomeMixer",
					   "GenomeMixer " + globalVersionStr + ", " + globalDateStr + "\n"
					   "Written in C++ by Alex Williams.\n"
					   "\n"
					   "Mac OS X, Unix, and Windows versions are available.\n"
					   "\n"
					   "See the online documentation at\n"
					   "http://www.nervenet.org/genome_mixer/\n");
}




void MixerApp::quickLoad() {
  // Loads in two files with the special names "quickmarkers.mar" and "quicksettings.gmx"
	try {
		QString theNewFileStr = "Saved_Settings/quicksettings.gmx";
		MouseIO::readBreedingSchemeFile(theNewFileStr, mainMixer);
		settingsNeedSaving = false;
		settingsLocation = theNewFileStr; // Now we should get the settings from this file
		textGeneralInfo->setText("Successfully loaded general settings file\n\"" 
								 + settingsLocation + "\"");
	} catch ( const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", QString(e->c_str()), "Ok");
		textGeneralInfo->setText("Failed to Quickload settings.\nQuickload looks for a settings file named quicksettings.gmx\nin the \"Saved_Settings\" folder (which must be in the folder where GenomeMixer\nis located). If no valid settings file with that name exists,\nthen Quickload will fail. If you don't have a Saved_Settings\nfolder, you can create a new one.");
		updateGeneralInfo();
		return;
	}
	
	try {
		QString theNewFileStr = "Saved_Settings/quickmarkers.mar";
		if (mainMixer->markerColPtr != NULL)  delete mainMixer->markerColPtr; // <--NOT "delete[ ]"
		mainMixer->markerColPtr = MouseIO::getMarkerDataFromFile(theNewFileStr);
		markersNeedSaving = false;	    
		markersLocation = theNewFileStr; // Now we should update the file location	    
		textGeneralInfo->setText("Successfully loaded marker file\n\"" + markersLocation + "\"");
		//mainMixer->markerColPtr->printMarkerInfo(); // <-- For debugging, prints what we just read in
		
	} catch ( const string *e) {
		QMessageBox::warning(this, "GenomeMixer Alert", "Failed to Quickload markers.\nQuickload looks for a markers file named quickmarkers.mar\nin the \"Saved_Settings\" folder (which must be in the folder where GenomeMixer\nis located). If no valid marker file with that name exists,\nthen Quickload will fail. If you don't have a Saved_Settings\nfolder, you can create a new one.", "Ok");
		textGeneralInfo->setText("Failed to Quickload the quickmarkers.mar file.\nPerhaps no such file exists in the Saved_Settings folder.");
		
		if (mainMixer->markerColPtr) {
			delete mainMixer->markerColPtr;
			mainMixer->markerColPtr = NULL;
		}
		
		updateGeneralInfo();
		return;
	}
	
	updateGeneralInfo();
}



void MixerApp::showInfoDialog() {
	theInfoDialog.show();
}
